@@ -32,7 +32,9 @@
 
 from __future__ import print_function
 
+import collections
 import datetime
+import io
 import json
 import os
 import pkg_resources
@@ -41,14 +43,23 @@ import shutil
 import sys
 import traceback
 
+# Python 2/3 support.
+try:
+    from configparser import SafeConfigParser
+except ImportError:
+    from ConfigParser import SafeConfigParser
 from dateutil import tz
 from pkg_resources import parse_version
 
 from bloom.generators import BloomGenerator
+from bloom.generators import GeneratorError
 from bloom.generators import resolve_dependencies
 from bloom.generators import update_rosdep
 
 from bloom.generators.common import default_fallback_resolver
+from bloom.generators.common import invalidate_view_cache
+from bloom.generators.common import evaluate_package_conditions
+from bloom.generators.common import resolve_rosdep_key
 
 from bloom.git import inbranch
 from bloom.git import get_branches
@@ -61,7 +72,6 @@ from bloom.git import tag_exists
 from bloom.logging import ansi
 from bloom.logging import debug
 from bloom.logging import enable_drop_first_log_prefix
-enable_drop_first_log_prefix(True)
 from bloom.logging import error
 from bloom.logging import fmt
 from bloom.logging import info
@@ -73,6 +83,7 @@ from bloom.commands.git.patch.common import set_patch_config
 
 from bloom.packages import get_package_data
 
+from bloom.util import code
 from bloom.util import to_unicode
 from bloom.util import execute_command
 from bloom.util import get_rfc_2822_date
@@ -83,7 +94,7 @@ try:
     from catkin_pkg.changelog import CHANGELOG_FILENAME
 except ImportError as err:
     debug(traceback.format_exc())
-    error("rosdep was not detected, please install it.", exit=True)
+    error("catkin_pkg was not detected, please install it.", exit=True)
 
 try:
     import rosdistro
@@ -108,6 +119,9 @@ except NameError:
     pass
 # End fix
 
+# Drop the first log prefix for this command
+enable_drop_first_log_prefix(True)
+
 TEMPLATE_EXTENSION = '.em'
 
 
@@ -134,16 +148,19 @@ def __place_template_folder(group, src, dst, gbp=False):
             if not os.path.exists(dst):
                 os.makedirs(dst)
             if os.path.exists(template_dst):
-                debug("Removing existing file '{0}'".format(template_dst))
-                os.remove(template_dst)
-            with open(template_dst, 'w') as f:
-                if not isinstance(template, str):
-                    template = template.decode('utf-8')
-                f.write(template)
-            shutil.copystat(template_abs_path, template_dst)
+                debug("Not overwriting existing file '{0}'".format(template_dst))
+            else:
+                with io.open(template_dst, 'w', encoding='utf-8') as f:
+                    if not isinstance(template, str):
+                        template = template.decode('utf-8')
+                    # Python 2 API needs a `unicode` not a utf-8 string.
+                    elif sys.version_info.major == 2:
+                        template = template.decode('utf-8')
+                    f.write(template)
+                shutil.copystat(template_abs_path, template_dst)
 
 
-def place_template_files(path, gbp=False):
+def place_template_files(path, build_type, gbp=False):
     info(fmt("@!@{bf}==>@| Placing templates files in the 'debian' folder."))
     debian_path = os.path.join(path, 'debian')
     # Create/Clean the debian folder
@@ -151,7 +168,8 @@ def place_template_files(path, gbp=False):
         os.makedirs(debian_path)
     # Place template files
     group = 'bloom.generators.debian'
-    __place_template_folder(group, 'templates', debian_path, gbp)
+    templates = os.path.join('templates', build_type)
+    __place_template_folder(group, templates, debian_path, gbp)
 
 
 def summarize_dependency_mapping(data, deps, build_deps, resolved_deps):
@@ -200,6 +218,29 @@ def format_depends(depends, resolved_deps):
     return formatted
 
 
+def format_description(value):
+    """
+    Format proper <synopsis, long desc> string following Debian control file
+    formatting rules. Treat first line in given string as synopsis, everything
+    else as a single, large paragraph.
+
+    Future extensions of this function could convert embedded newlines and / or
+    html into paragraphs in the Description field.
+
+    https://www.debian.org/doc/debian-policy/ch-controlfields.html#s-f-Description
+    """
+    value = debianize_string(value)
+    # NOTE: bit naive, only works for 'properly formatted' pkg descriptions (ie:
+    #       'Text. Text'). Extra space to avoid splitting on arbitrary sequences
+    #       of characters broken up by dots (version nrs fi).
+    parts = value.split('. ', 1)
+    if len(parts) == 1 or len(parts[1]) == 0:
+        # most likely single line description
+        return value
+    # format according to rules in linked field documentation
+    return u"{0}.\n {1}".format(parts[0], parts[1].strip())
+
+
 def get_changelogs(package, releaser_history=None):
     if releaser_history is None:
         warning("No historical releaser history, using current maintainer name "
@@ -248,14 +289,15 @@ def generate_substitutions_from_package(
     deb_inc=0,
     peer_packages=None,
     releaser_history=None,
-    fallback_resolver=None
+    fallback_resolver=None,
+    native=False
 ):
     peer_packages = peer_packages or []
     data = {}
     # Name, Version, Description
     data['Name'] = package.name
     data['Version'] = package.version
-    data['Description'] = debianize_string(package.description)
+    data['Description'] = format_description(package.description)
     # Websites
     websites = [str(url) for url in package.urls if url.type == 'website']
     homepage = websites[0] if websites else ''
@@ -263,15 +305,26 @@ def generate_substitutions_from_package(
         warning("No homepage set, defaulting to ''")
     data['Homepage'] = homepage
     # Debian Increment Number
-    data['DebianInc'] = deb_inc
+    data['DebianInc'] = '' if native else '-{0}'.format(deb_inc)
+    # Debian Package Format
+    data['format'] = 'native' if native else 'quilt'
     # Package name
     data['Package'] = sanitize_package_name(package.name)
     # Installation prefix
     data['InstallationPrefix'] = installation_prefix
     # Resolve dependencies
-    depends = package.run_depends + package.buildtool_export_depends
-    build_depends = package.build_depends + package.buildtool_depends + package.test_depends
-    unresolved_keys = depends + build_depends + package.replaces + package.conflicts
+    evaluate_package_conditions(package, ros_distro)
+    depends = [
+        dep for dep in (package.run_depends + package.buildtool_export_depends)
+        if dep.evaluated_condition is not False]
+    build_depends = [
+        dep for dep in (package.build_depends + package.buildtool_depends + package.test_depends)
+        if dep.evaluated_condition is not False]
+
+    unresolved_keys = [
+        dep for dep in (depends + build_depends + package.replaces + package.conflicts)
+        if dep.evaluated_condition is not False]
+    # The installer key is not considered here, but it is checked when the keys are checked before this
     resolved_deps = resolve_dependencies(unresolved_keys, os_name,
                                          os_version, ros_distro,
                                          peer_packages + [d.name for d in package.replaces + package.conflicts],
@@ -288,6 +341,33 @@ def generate_substitutions_from_package(
     data['Conflicts'] = sorted(
         set(format_depends(package.conflicts, resolved_deps))
     )
+
+    # Build-type specific substitutions.
+    build_type = package.get_build_type()
+    if build_type == 'catkin':
+        pass
+    elif build_type == 'cmake':
+        pass
+    elif build_type == 'ament_cmake':
+        pass
+    elif build_type == 'ament_python':
+        # Don't set the install-scripts flag if it's already set in setup.cfg.
+        package_path = os.path.abspath(os.path.dirname(package.filename))
+        setup_cfg_path = os.path.join(package_path, 'setup.cfg')
+        data['pass_install_scripts'] = True
+        if os.path.isfile(setup_cfg_path):
+            setup_cfg = SafeConfigParser()
+            setup_cfg.read([setup_cfg_path])
+            if (
+                    setup_cfg.has_option('install', 'install-scripts') or
+                    setup_cfg.has_option('install', 'install_scripts')
+            ):
+                data['pass_install_scripts'] = False
+    else:
+        error(
+            "Build type '{}' is not supported by this version of bloom.".
+            format(build_type), exit=True)
+
     # Set the distribution
     data['Distribution'] = os_version
     # Use the time stamp to set the date strings
@@ -345,6 +425,20 @@ def generate_substitutions_from_package(
     data['debhelper_version'] = 7 if os_version in ['oneiric'] else 9
     # Summarize dependencies
     summarize_dependency_mapping(data, depends, build_depends, resolved_deps)
+    # Copyright
+    licenses = []
+    separator = '\n' + '=' * 80 + '\n\n'
+    for l in package.licenses:
+        if hasattr(l, 'file') and l.file is not None:
+            license_file = os.path.join(os.path.dirname(package.filename), l.file)
+            if not os.path.exists(license_file):
+                error("License file '{}' is not found.".
+                      format(license_file), exit=True)
+            license_text = open(license_file, 'r').read()
+            if not license_text.endswith('\n'):
+                license_text += '\n'
+            licenses.append(license_text)
+    data['Copyright'] = separator.join(licenses)
 
     def convertToUnicode(obj):
         if sys.version_info.major == 2:
@@ -399,8 +493,15 @@ def __process_template_folder(path, subs):
             os.path.relpath(item),
             os.path.relpath(template_path)))
         result = em.expand(template, **subs)
+        # Don't write an empty file
+        if len(result) == 0 and \
+           os.path.basename(template_path) in ['copyright']:
+            processed_items.append(item)
+            continue
         # Write the result
-        with open(template_path, 'w') as f:
+        with io.open(template_path, 'w', encoding='utf-8') as f:
+            if sys.version_info.major == 2:
+                result = result.decode('utf-8')
             f.write(result)
         # Copy the permissions
         shutil.copymode(item, template_path)
@@ -474,9 +575,9 @@ def sanitize_package_name(name):
 class DebianGenerator(BloomGenerator):
     title = 'debian'
     description = "Generates debians from the catkin meta data"
-    has_run_rosdep = False
+    has_run_rosdep = os.environ.get('BLOOM_SKIP_ROSDEP_UPDATE', '0').lower() not in ['0', 'f', 'false', 'n', 'no']
     default_install_prefix = '/usr'
-    rosdistro = os.environ.get('ROS_DISTRO', 'groovy')
+    rosdistro = os.environ.get('ROS_DISTRO', 'indigo')
 
     def prepare_arguments(self, parser):
         # Add command line arguments for this generator
@@ -494,6 +595,9 @@ class DebianGenerator(BloomGenerator):
             help="overrides the default installation prefix (/usr)")
         add('--os-name', default='ubuntu',
             help="overrides os_name, set to 'ubuntu' by default")
+        add('--os-not-required', default=False, action="store_true",
+            help="Do not error if this os is not in the platforms "
+                 "list for rosdistro")
 
     def handle_arguments(self, args):
         self.interactive = args.interactive
@@ -504,6 +608,11 @@ class DebianGenerator(BloomGenerator):
             index = rosdistro.get_index(rosdistro.get_index_url())
             distribution_file = rosdistro.get_distribution_file(index, self.rosdistro)
             if self.os_name not in distribution_file.release_platforms:
+                if args.os_not_required:
+                    warning("No platforms defined for os '{0}' in release file for the "
+                            "'{1}' distro. This os was not required; continuing without error."
+                            .format(self.os_name, self.rosdistro))
+                    sys.exit(0)
                 error("No platforms defined for os '{0}' in release file for the '{1}' distro."
                       .format(self.os_name, self.rosdistro), exit=True)
             self.distros = distribution_file.release_platforms[self.os_name]
@@ -546,6 +655,59 @@ class DebianGenerator(BloomGenerator):
         update_rosdep()
         self.has_run_rosdep = True
 
+    def _check_all_keys_are_valid(self, peer_packages, ros_distro):
+        keys_to_resolve = []
+        key_to_packages_which_depends_on = collections.defaultdict(list)
+        keys_to_ignore = set()
+        for package in self.packages.values():
+            evaluate_package_conditions(package, ros_distro)
+            depends = [
+                dep for dep in (package.run_depends + package.buildtool_export_depends)
+                if dep.evaluated_condition is not False]
+            build_depends = [
+                dep for dep in (package.build_depends + package.buildtool_depends + package.test_depends)
+                if dep.evaluated_condition is not False]
+            unresolved_keys = [
+                dep for dep in (depends + build_depends + package.replaces + package.conflicts)
+                if dep.evaluated_condition is not False]
+            keys_to_ignore = {
+                    dep for dep in keys_to_ignore.union(package.replaces + package.conflicts)
+                    if dep.evaluated_condition is not False}
+            keys = [d.name for d in unresolved_keys]
+            keys_to_resolve.extend(keys)
+            for key in keys:
+                key_to_packages_which_depends_on[key].append(package.name)
+
+        os_name = self.os_name
+        rosdistro = self.rosdistro
+        all_keys_valid = True
+        for key in sorted(set(keys_to_resolve)):
+            for os_version in self.distros:
+                try:
+                    extended_peer_packages = peer_packages + [d.name for d in keys_to_ignore]
+                    rule, installer_key, default_installer_key = \
+                        resolve_rosdep_key(key, os_name, os_version, rosdistro, extended_peer_packages,
+                                           retry=False)
+                    if rule is None:
+                        continue
+                    if installer_key != default_installer_key:
+                        error("Key '{0}' resolved to '{1}' with installer '{2}', "
+                              "which does not match the default installer '{3}'."
+                              .format(key, rule, installer_key, default_installer_key))
+                        BloomGenerator.exit(
+                            "The Debian generator does not support dependencies "
+                            "which are installed with the '{0}' installer."
+                            .format(installer_key),
+                            returncode=code.GENERATOR_INVALID_INSTALLER_KEY)
+                except (GeneratorError, RuntimeError) as e:
+                    print(fmt("Failed to resolve @{cf}@!{key}@| on @{bf}{os_name}@|:@{cf}@!{os_version}@| with: {e}")
+                          .format(**locals()))
+                    print(fmt("@{cf}@!{0}@| is depended on by these packages: ").format(key) +
+                          str(list(set(key_to_packages_which_depends_on[key]))))
+                    print(fmt("@{kf}@!<== @{rf}@!Failed@|"))
+                    all_keys_valid = False
+        return all_keys_valid
+
     def pre_modify(self):
         info("\nPre-verifying Debian dependency keys...")
         # Run rosdep update is needed
@@ -554,15 +716,18 @@ class DebianGenerator(BloomGenerator):
 
         peer_packages = [p.name for p in self.packages.values()]
 
-        for package in self.packages.values():
-            depends = package.run_depends + package.buildtool_export_depends
-            build_depends = package.build_depends + package.buildtool_depends + package.test_depends
-            unresolved_keys = depends + build_depends + package.replaces + package.conflicts
-            for os_version in self.distros:
-                resolve_dependencies(unresolved_keys, self.os_name,
-                                     os_version, self.rosdistro,
-                                     peer_packages + [d.name for d in package.replaces + package.conflicts],
-                                     fallback_resolver=missing_dep_resolver)
+        while not self._check_all_keys_are_valid(peer_packages, self.rosdistro):
+            error("Some of the dependencies for packages in this repository could not be resolved by rosdep.",
+                  exit=not self.interactive)
+            error("You can try to address the issues which appear above and try again if you wish.")
+            try:
+                if not maybe_continue(msg="Would you like to try again?"):
+                    error("User aborted after rosdep keys were not resolved.")
+                    sys.exit(code.GENERATOR_NO_ROSDEP_KEY_FOR_DISTRO)
+            except (KeyboardInterrupt, EOFError):
+                error("\nUser quit.", exit=True)
+            update_rosdep()
+            invalidate_view_cache()
 
         info("All keys are " + ansi('greenf') + "OK" + ansi('reset') + "\n")
 
@@ -599,7 +764,7 @@ class DebianGenerator(BloomGenerator):
                  .format(destination))
             # Then this is a debian branch
             # Place the raw template files
-            self.place_template_files()
+            self.place_template_files(package.get_build_type())
         else:
             # This is a distro specific debian branch
             # Determine the current package being generated
@@ -671,7 +836,7 @@ class DebianGenerator(BloomGenerator):
             return config_store
         return json.loads(config_store)
 
-    def place_template_files(self, debian_dir='debian'):
+    def place_template_files(self, build_type, debian_dir='debian'):
         # Create/Clean the debian folder
         if os.path.exists(debian_dir):
             if self.interactive:
@@ -688,10 +853,12 @@ class DebianGenerator(BloomGenerator):
             else:
                 warning("Not overwriting debian directory.")
         # Use generic place template files command
-        place_template_files('.', gbp=True)
+        place_template_files('.', build_type, gbp=True)
         # Commit results
         execute_command('git add ' + debian_dir)
-        execute_command('git commit -m "Placing debian template files"')
+        _, has_files, _ = execute_command('git diff --cached --name-only', return_io=True)
+        if has_files:
+            execute_command('git commit -m "Placing debian template files"')
 
     def get_releaser_history(self):
         # Assumes that this is called in the target branch
@@ -737,7 +904,7 @@ class DebianGenerator(BloomGenerator):
         # Template files
         template_files = process_template_files('.', subs)
         # Remove any residual template files
-        execute_command('git rm -rf ' + ' '.join(template_files))
+        execute_command('git rm -rf ' + ' '.join("'{}'".format(t) for t in template_files))
         # Add changes to the debian folder
         execute_command('git add debian')
         # Commit changes
@@ -751,7 +918,7 @@ class DebianGenerator(BloomGenerator):
                                             self.debian_inc)
 
     def generate_tag_name(self, data):
-        tag_name = '{Package}_{Version}-{DebianInc}_{Distribution}'
+        tag_name = '{Package}_{Version}{DebianInc}_{Distribution}'
         tag_name = 'debian/' + tag_name.format(**data)
         return tag_name
 
