@@ -32,7 +32,9 @@
 
 from __future__ import print_function
 
+import collections
 import datetime
+import io
 import json
 import os
 import pkg_resources
@@ -42,14 +44,24 @@ import sys
 import traceback
 import textwrap
 
+# Python 2/3 support.
+try:
+    from configparser import SafeConfigParser
+except ImportError:
+    from ConfigParser import SafeConfigParser
 from dateutil import tz
+from distutils.version import LooseVersion
 from time import strptime
 
 from bloom.generators import BloomGenerator
+from bloom.generators import GeneratorError
 from bloom.generators import resolve_dependencies
 from bloom.generators import update_rosdep
 
 from bloom.generators.common import default_fallback_resolver
+from bloom.generators.common import invalidate_view_cache
+from bloom.generators.common import evaluate_package_conditions
+from bloom.generators.common import resolve_rosdep_key
 
 from bloom.git import inbranch
 from bloom.git import get_branches
@@ -62,11 +74,9 @@ from bloom.git import tag_exists
 from bloom.logging import ansi
 from bloom.logging import debug
 from bloom.logging import enable_drop_first_log_prefix
-enable_drop_first_log_prefix(True)
 from bloom.logging import error
 from bloom.logging import fmt
 from bloom.logging import info
-from bloom.logging import is_debug
 from bloom.logging import warning
 
 from bloom.commands.git.patch.common import get_patch_config
@@ -74,17 +84,10 @@ from bloom.commands.git.patch.common import set_patch_config
 
 from bloom.packages import get_package_data
 
-from bloom.util import to_unicode
+from bloom.util import code
 from bloom.util import execute_command
 from bloom.util import maybe_continue
 
-try:
-    from catkin_pkg.changelog import get_changelog_from_path
-    from catkin_pkg.changelog import CHANGELOG_FILENAME
-except ImportError as err:
-    debug(traceback.format_exc())
-    error("rosdep was not detected, please install it.", exit=True)
-
 try:
     import rosdistro
 except ImportError as err:
@@ -97,6 +100,9 @@ except ImportError:
     debug(traceback.format_exc())
     error("empy was not detected, please install it.", exit=True)
 
+# Drop the first log prefix for this command
+enable_drop_first_log_prefix(True)
+
 TEMPLATE_EXTENSION = '.em'
 
 
@@ -129,7 +135,7 @@ def __place_template_folder(group, src, dst, gbp=False):
             shutil.copystat(template_abs_path, template_dst)
 
 
-def place_template_files(path, gbp=False):
+def place_template_files(path, build_type, gbp=False):
     info(fmt("@!@{bf}==>@| Placing templates files in the 'rpm' folder."))
     rpm_path = os.path.join(path, 'rpm')
     # Create/Clean the rpm folder
@@ -137,7 +143,8 @@ def place_template_files(path, gbp=False):
         os.makedirs(rpm_path)
     # Place template files
     group = 'bloom.generators.rpm'
-    __place_template_folder(group, 'templates', rpm_path, gbp)
+    templates = os.path.join('templates', build_type)
+    __place_template_folder(group, templates, rpm_path, gbp)
 
 
 def summarize_dependency_mapping(data, deps, build_deps, resolved_deps):
@@ -145,8 +152,8 @@ def summarize_dependency_mapping(data, deps, build_deps, resolved_deps):
         return
     info("Package '" + data['Package'] + "' has dependencies:")
     header = "  " + ansi('boldoff') + ansi('ulon') + \
-             "rosdep key           => " + data['Distribution'] + \
-             " key" + ansi('reset')
+             "rosdep key           => " + data['OSName'] + ' ' + \
+             data['Distribution'] + " key" + ansi('reset')
     template = "  " + ansi('cyanf') + "{0:<20} " + ansi('purplef') + \
                "=> " + ansi('cyanf') + "{1}" + ansi('reset')
     if len(deps) != 0:
@@ -225,9 +232,17 @@ def generate_substitutions_from_package(
     # Installation prefix
     data['InstallationPrefix'] = installation_prefix
     # Resolve dependencies
-    depends = package.run_depends + package.buildtool_export_depends
-    build_depends = package.build_depends + package.buildtool_depends + package.test_depends
-    unresolved_keys = depends + build_depends + package.replaces + package.conflicts
+    evaluate_package_conditions(package, ros_distro)
+    depends = [
+        dep for dep in (package.run_depends + package.buildtool_export_depends)
+        if dep.evaluated_condition is not False]
+    build_depends = [
+        dep for dep in (package.build_depends + package.buildtool_depends + package.test_depends)
+        if dep.evaluated_condition is not False]
+    unresolved_keys = [
+        dep for dep in (depends + build_depends + package.replaces + package.conflicts)
+        if dep.evaluated_condition is not False]
+    # The installer key is not considered here, but it is checked when the keys are checked before this
     resolved_deps = resolve_dependencies(unresolved_keys, os_name,
                                          os_version, ros_distro,
                                          peer_packages + [d.name for d in package.replaces + package.conflicts],
@@ -244,7 +259,26 @@ def generate_substitutions_from_package(
     data['Conflicts'] = sorted(
         set(format_depends(package.conflicts, resolved_deps))
     )
-    # Set the distribution
+    data['Provides'] = []
+    data['Supplements'] = []
+
+    # Build-type specific substitutions.
+    build_type = package.get_build_type()
+    if build_type == 'catkin':
+        pass
+    elif build_type == 'cmake':
+        pass
+    elif build_type == 'ament_cmake':
+        pass
+    elif build_type == 'ament_python':
+        pass
+    else:
+        error(
+            "Build type '{}' is not supported by this version of bloom.".
+            format(build_type), exit=True)
+
+    # Set the OS and distribution
+    data['OSName'] = os_name
     data['Distribution'] = os_version
     # Use the time stamp to set the date strings
     stamp = datetime.datetime.now(tz.tzlocal())
@@ -258,6 +292,8 @@ def generate_substitutions_from_package(
     # Changelog
     if releaser_history:
         sorted_releaser_history = sorted(releaser_history,
+                                         key=lambda k: LooseVersion(k), reverse=True)
+        sorted_releaser_history = sorted(sorted_releaser_history,
                                          key=lambda k: strptime(releaser_history.get(k)[0], '%a %b %d %Y'),
                                          reverse=True)
         changelogs = [(v, releaser_history[v]) for v in sorted_releaser_history]
@@ -334,7 +370,9 @@ def __process_template_folder(path, subs):
             os.path.relpath(template_path)))
         result = em.expand(template, **subs)
         # Write the result
-        with open(template_path, 'w') as f:
+        with io.open(template_path, 'w', encoding='utf-8') as f:
+            if sys.version_info.major == 2:
+                result = result.decode('utf-8')
             f.write(result)
         # Copy the permissions
         shutil.copymode(item, template_path)
@@ -409,9 +447,9 @@ def sanitize_package_name(name):
 class RpmGenerator(BloomGenerator):
     title = 'rpm'
     description = "Generates RPMs from the catkin meta data"
-    has_run_rosdep = False
+    has_run_rosdep = os.environ.get('BLOOM_SKIP_ROSDEP_UPDATE', '0').lower() not in ['0', 'f', 'false', 'n', 'no']
     default_install_prefix = '/usr'
-    rosdistro = os.environ.get('ROS_DISTRO', 'groovy')
+    rosdistro = os.environ.get('ROS_DISTRO', 'indigo')
 
     def prepare_arguments(self, parser):
         # Add command line arguments for this generator
@@ -474,6 +512,7 @@ class RpmGenerator(BloomGenerator):
     def summarize(self):
         info("Generating source RPMs for the packages: " + str(self.names))
         info("RPM Incremental Version: " + str(self.rpm_inc))
+        info("RPM OS: " + str(self.os_name))
         info("RPM Distributions: " + str(self.distros))
 
     def get_branching_arguments(self):
@@ -483,6 +522,59 @@ class RpmGenerator(BloomGenerator):
         update_rosdep()
         self.has_run_rosdep = True
 
+    def _check_all_keys_are_valid(self, peer_packages, rosdistro):
+        keys_to_resolve = []
+        key_to_packages_which_depends_on = collections.defaultdict(list)
+        keys_to_ignore = set()
+        for package in self.packages.values():
+            evaluate_package_conditions(package, rosdistro)
+            depends = [
+                dep for dep in (package.run_depends + package.buildtool_export_depends)
+                if dep.evaluated_condition is not False]
+            build_depends = [
+                dep for dep in (package.build_depends + package.buildtool_depends + package.test_depends)
+                if dep.evaluated_condition is not False]
+            unresolved_keys = [
+                dep for dep in (depends + build_depends + package.replaces + package.conflicts)
+                if dep.evaluated_condition is not False]
+            keys_to_ignore = {
+                dep for dep in keys_to_ignore.union(package.replaces + package.conflicts)
+                if dep.evaluated_condition is not False}
+            keys = [d.name for d in unresolved_keys]
+            keys_to_resolve.extend(keys)
+            for key in keys:
+                key_to_packages_which_depends_on[key].append(package.name)
+
+        os_name = self.os_name
+        rosdistro = self.rosdistro
+        all_keys_valid = True
+        for key in sorted(set(keys_to_resolve)):
+            for os_version in self.distros:
+                try:
+                    extended_peer_packages = peer_packages + [d.name for d in keys_to_ignore]
+                    rule, installer_key, default_installer_key = \
+                        resolve_rosdep_key(key, os_name, os_version, rosdistro, extended_peer_packages,
+                                           retry=False)
+                    if rule is None:
+                        continue
+                    if installer_key != default_installer_key:
+                        error("Key '{0}' resolved to '{1}' with installer '{2}', "
+                              "which does not match the default installer '{3}'."
+                              .format(key, rule, installer_key, default_installer_key))
+                        BloomGenerator.exit(
+                            "The RPM generator does not support dependencies "
+                            "which are installed with the '{0}' installer."
+                            .format(installer_key),
+                            returncode=code.GENERATOR_INVALID_INSTALLER_KEY)
+                except (GeneratorError, RuntimeError) as e:
+                    print(fmt("Failed to resolve @{cf}@!{key}@| on @{bf}{os_name}@|:@{cf}@!{os_version}@| with: {e}")
+                          .format(**locals()))
+                    print(fmt("@{cf}@!{0}@| is depended on by these packages: ").format(key) +
+                          str(list(set(key_to_packages_which_depends_on[key]))))
+                    print(fmt("@{kf}@!<== @{rf}@!Failed@|"))
+                    all_keys_valid = False
+        return all_keys_valid
+
     def pre_modify(self):
         info("\nPre-verifying RPM dependency keys...")
         # Run rosdep update is needed
@@ -491,15 +583,19 @@ class RpmGenerator(BloomGenerator):
 
         peer_packages = [p.name for p in self.packages.values()]
 
-        for package in self.packages.values():
-            depends = package.run_depends + package.buildtool_export_depends
-            build_depends = package.build_depends + package.buildtool_depends + package.test_depends
-            unresolved_keys = depends + build_depends + package.replaces + package.conflicts
-            for os_version in self.distros:
-                resolve_dependencies(unresolved_keys, self.os_name,
-                                     os_version, self.rosdistro,
-                                     peer_packages + [d.name for d in package.replaces + package.conflicts],
-                                     fallback_resolver=missing_dep_resolver)
+        while not self._check_all_keys_are_valid(peer_packages, self.rosdistro):
+            error("Some of the dependencies for packages in this repository could not be resolved by rosdep.",
+                  exit=not self.interactive)
+            error("You can try to address the issues which appear above and try again if you wish, "
+                  "or continue without releasing into RPM-based distributions (e.g. Fedora 24).")
+            try:
+                if not maybe_continue(msg="Would you like to try again?"):
+                    error("User aborted after rosdep keys were not resolved.")
+                    sys.exit(code.GENERATOR_NO_ROSDEP_KEY_FOR_DISTRO)
+            except (KeyboardInterrupt, EOFError):
+                error("\nUser quit.", exit=True)
+            update_rosdep()
+            invalidate_view_cache()
 
         info("All keys are " + ansi('greenf') + "OK" + ansi('reset') + "\n")
 
@@ -540,7 +636,7 @@ class RpmGenerator(BloomGenerator):
                  .format(destination))
             # Then this is an rpm branch
             # Place the raw template files
-            self.place_template_files()
+            self.place_template_files(package.get_build_type())
         else:
             # This is a distro specific rpm branch
             # Determine the current package being generated
@@ -588,8 +684,8 @@ class RpmGenerator(BloomGenerator):
         info(ansi(color) + "####" + ansi('reset'), use_prefix=False)
         info(
             ansi(color) + "#### " + ansi('greenf') + "Successfully" +
-            ansi(color) + " generated '" + ansi('boldon') + distro +
-            ansi('boldoff') + "' RPM for package"
+            ansi(color) + " generated '" + ansi('boldon') + self.os_name +
+            ' ' + distro + ansi('boldoff') + "' RPM for package"
             " '" + ansi('boldon') + package.name + ansi('boldoff') + "'" +
             " at version '" + ansi('boldon') + package.version +
             "-" + str(self.rpm_inc) + ansi('boldoff') + "'" +
@@ -612,7 +708,7 @@ class RpmGenerator(BloomGenerator):
             return config_store
         return json.loads(config_store)
 
-    def place_template_files(self, rpm_dir='rpm'):
+    def place_template_files(self, build_type, rpm_dir='rpm'):
         # Create/Clean the rpm folder
         if os.path.exists(rpm_dir):
             if self.interactive:
@@ -627,7 +723,7 @@ class RpmGenerator(BloomGenerator):
             if os.path.exists(rpm_dir):
                 shutil.rmtree(rpm_dir)
         # Use generic place template files command
-        place_template_files('.', gbp=True)
+        place_template_files('.', build_type, gbp=True)
         # Commit results
         execute_command('git add ' + rpm_dir)
         execute_command('git commit -m "Placing rpm template files"')
@@ -663,7 +759,7 @@ class RpmGenerator(BloomGenerator):
         )
 
     def generate_rpm(self, package, rpm_distro, rpm_dir='rpm'):
-        info("Generating RPM for {0}...".format(rpm_distro))
+        info("Generating RPM for {0} {1}...".format(self.os_name, rpm_distro))
         # Try to retrieve the releaser_history
         releaser_history = self.get_releaser_history()
         # Generate substitution values
@@ -673,9 +769,11 @@ class RpmGenerator(BloomGenerator):
         # Template files
         template_files = process_template_files('.', subs)
         # Remove any residual template files
-        execute_command('git rm -rf ' + ' '.join(template_files))
-        # Add changes to the rpm folder
-        execute_command('git add ' + rpm_dir)
+        execute_command('git rm -rf ' + ' '.join("'{}'".format(t) for t in template_files))
+        # Add marker file to tell mock to archive the sources
+        open('.write_tar', 'a').close()
+        # Add marker file changes to the rpm folder
+        execute_command('git add .write_tar ' + rpm_dir)
         # Commit changes
         execute_command('git commit -m "Generated RPM files for ' +
                         rpm_distro + '"')
@@ -707,8 +805,8 @@ class RpmGenerator(BloomGenerator):
     def summarize_package(self, package, distro, color='bluef'):
         info(ansi(color) + "\n####" + ansi('reset'), use_prefix=False)
         info(
-            ansi(color) + "#### Generating '" + ansi('boldon') + distro +
-            ansi('boldoff') + "' RPM for package"
+            ansi(color) + "#### Generating '" + ansi('boldon') + self.os_name +
+            ' ' + distro + ansi('boldoff') + "' RPM for package"
             " '" + ansi('boldon') + package.name + ansi('boldoff') + "'" +
             " at version '" + ansi('boldon') + package.version +
             "-" + str(self.rpm_inc) + ansi('boldoff') + "'" +
